#!/usr/bin/perl -w
#
# Copyright (c) 2000 Easysoft Ltd. All rights reserved.
#
# This Perl script reads a Washington University FTP log, creates a table
# in a database and inserts information from the FTP log. This script has
# been run through OOB to MS SQLServer, MS Access and Easysoft's own database.
# If you use a different database you may need to modify some of the SQL
# syntax (e.g. the grant and insert).
#
# $Id: ftp_downloads.pl 91 2004-05-20 15:05:37Z martin $
#
# $Log: ftp_downloads.pl,v $
# Revision 1.2  2004/04/14 14:58:30  martin
# test on opts{i} wrong way around.
#
# Revision 1.1  2000/03/09 14:22:05  martin
# Initial revision
#
#
# usage:
#
# ftp_downloads.pl -d dsn -u dbuser -p dbpassword {-t table} 
#                  -i ignore_this_host < xferlog
#
# where:
#
# dsn is the datasource name you must define in your odbc.ini. Just specify
# the dsn name and not the full DBI name, i.e. -d test not -d dbi:ODBC:test
#
# dbuser is the database username. If your database does not require this
# then omit -u.
#
# dbpassword is the database password for dbuser. If your database does
# not require a username and password omit -p.
#
# table is the name of the table to create and insert records in to.
#
# ignore_this_host is a string/substring of a host entry. If the FTP host
# column contains this string the record is ignored.
#
# xferlog is your Washington University FTP log.
# A typical entry looks like this:
#
# Tue Jan 25 14:50:54 2000 1 devmaster.easysoft.com 3571308 
# /pub/EasysoftODBC-ODBCBridge1_0_0_0.exe b _ o g rftp ftp 0 *
#
# The columns are:
#
#	  1 current time in the form DDD MMM dd hh:mm:ss YYYY
#	  2 transfer time in seconds
#	  3 remote host name
#	  4 file size in bytes
#	  5 name of file
#	  6 transfer type (a>scii, b>inary)
#	  7 special action flags (concatenated as needed):
#		C   file was compressed
#		U   file was uncompressed
#		T   file was tar'ed
#		_   no action taken
#	  8 file was sent to user (o>utgoing) or received from
#	    user (i>ncoming)
#	  9 accessed anonymously (r>eal, a>nonymous, g>uest) -- mostly for FTP
#	 10 local username or, if guest, ID string given
#	    (anonymous FTP password)
#	 11 service name ('ftp', other)
#	 12 authentication method (bitmask)
#		0   none
#		1   RFC931 Authentication
#	 13 authenticated user id (if available, '*' otherwise)
#
use IO::Handle;
use DBI;
use DBI qw(:sql_types);
use Getopt::Std;

my $dsn="dbi:ODBC:DSN=";	# DSN name, append -d value later
my $db_user="";			# database username
my $db_auth="";			# database password
my $table="";
my $dotper=100;
my $usage= "Usage:\n\nftp_downloads.pl -d datasource_name " .
    "{-u database_username} {-p database_password} {-t table_name} " .
    "-i ignore_host\n";

#
# Parse the command line and check.
#
my %opts = ();

getopts('d:u:p:t:i:', \%opts) || die $usage;

if (!exists($opts{d}))
{
    print "Invalid command line\n\n";
    print $usage;
    exit 1;
}
if (!exists($opts{u}))
{
    $db_user="";
}
else
{
    $db_user = $opts{u};
}
if (!exists($opts{p}))
{
    $db_auth="";
}
else
{
    $db_auth = $opts{p};
}
if (!exists($opts{t}))
{ 
    $table = "ftp_downloads";
}
else
{
    $table = $opts{t};
}

$dsn .= $opts{d};

print "Using DSN : ", $dsn, "\n";
print "Using DB username : ", $db_user, "\n";
print "Using DB password : ", $db_auth, "\n";
print "Using table : ", $table, "\n";
if (exists($opts{i}))
{
    print "Ignoring entries from hosts : ", $opts{i}, "\n";
}
#
# Connect to the data source and do all the hard work.
#
my $dbh = DBI->connect($dsn,
		       $db_user,
		       $db_auth,
		       {
			   RaiseError => 0,
			   AutoCommit => 1
                           }
		       ) || die "Database connection failed: $DBI::errstr";
my $sql;

print "NOTE: This example will not work correctly with wu.ftpd logs which\n";
print "      have filenames containing spaces.\n";
#
# Delete the table so we can start again.
#
print "Dropping table $table\n";
$dbh->do("drop table $table");
#
# Create the table.
#
print "Creating Table $table\n";
$dbh->do("Create table $table (dt datetime,downloadtime integer,host varchar(100),bytes integer,path varchar(255),transfertype char(1),usertype char(1),email varchar(100))");
#
# Grant everyone access to the table.
#
print "Granting everyone access to the table $table\n";
$dbh->do("grant all on $table to public");
#
# Read the ftp log and insert the data into the table.
#
my $inserts = 0;
#
# We want to see the dots representing dotper insertions immediately.
#
open(DEV, ">/dev/tty");
DEV->autoflush(1);

while(<>)
{
    my ($dayname, $month, $daynum, $time, $year, $downloadtime, $hostname);
    my ($bytes, $path, $transfertype, $ig1, $ig2, $usertype, $mail);
    my $the_rest;
    my %mon_to_num=("Jan", "01", "Feb", "02", "Mar", "03", "Apr", "04",
	"May", "05", "Jun", "06", "Jul", "07", "Aug", "08", "Sep", "09",
	"Oct", "10", "Nov", "11", "Dec", "12");
    my @arr;

    ($dayname,$month,$daynum,$time,$year,$downloadtime,$hostname,$bytes,$path,
     $transfertype,$ig1,$ig2,$usertype,$mail,$the_rest) = (split);
    @arr = (split);
    if (!($transfertype =~ /^a$/) and  !($transfertype =~ "b")){
        #
        # uncomment the following to see which records are ignored.
	#print "Ignoring:\n", @arr, "\n";
    	next;
    }
    next if (exists($opts{i}) && ($hostname =~ /$opts{i}/));
    #
    # MS SQLServer insists on 2 digit days.
    #
    ($daynum < 10) and $daynum = "0$daynum";
    #
    # Quote any possible special characers in path and mail address.
    #
    $path=$dbh->quote($path);
    $mail=$dbh->quote($mail);
    #
    # Form the SQL to send to the database.
    #
    $sql = "insert into $table values(" .
    "{ts'$year-$mon_to_num{$month}-$daynum $time'}, " .
	"$downloadtime,'$hostname',$bytes,$path,'$transfertype'," .
	    "'$usertype',$mail)";
    #print $sql, "\n";
    $dbh->do($sql);
    $inserts++;
    if (($inserts % $dotper) == 0)
    {
	print DEV ".";
    }
}
print "\n", $inserts, " records inserted in to ", $table, "\n";
$dbh->disconnect;
